/*
Copyright 2018, 2019 eomanis

This file is part of getgarfield.

getgarfield is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3 as
published by the Free Software Foundation.

getgarfield is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with getgarfield.  If not, see <http://www.gnu.org/licenses/>.
*/

package getgarfield;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.SeekableByteChannel;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.nio.file.StandardOpenOption;
import java.time.LocalDate;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.Iterator;
import java.util.Locale;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Stream;

/**
 * <p>
 * Downloads all Garfield comic strips into a given target directory.
 * </p>
 * <p>
 * Also creates two web sites that present the comic strips by month, one web site that displays the downloaded strips and another that uses the original source
 * URLs where the strips have been downloaded from.
 * </p>
 * <p>
 * Only downloads comic strips that have not been downloaded already.
 * </p>
 * <p>
 * As of December 2018 the downloads amount to about 1.7 GiB.
 * </p>
 *
 * @author eomanis
 */
public class GetGarfield implements Runnable {

	public static final String APPLICATION_NAME = "getgarfield";
	public static final String APPLICATION_DESCRIPTION = "Java application that downloads all Garfield comic strips";

	// Semantic versioning
	public static final int VERSION_MAJOR = 0;
	public static final int VERSION_MINOR = 1;
	public static final int VERSION_PATCH = 0;
	public static final String VERSION_LABEL = "";
	public static final String VERSION = VERSION_MAJOR + "." + VERSION_MINOR + "." + VERSION_PATCH
			+ (!VERSION_LABEL.isEmpty() ? ("-" + VERSION_LABEL) : "");

	public static final LocalDate FIRST_COMIC_STRIP_DATE = LocalDate.parse( "1978-06-19" );

	private static final LocalDate LATEST_COMIC_STRIP_DATE_KNOWN = LocalDate.parse( "2019-01-01" );
	private static final Locale LOCALE_DATE_TIME_FORMATTERS = new Locale( "en" );
	private static final ZoneId TIME_ZONE_COMIC_STRIP = ZoneId.of( "US/Eastern" );
	private static final int CONCURRENT_DOWNLOADS = 8;
	private static final int LATEST_COMIC_STRIP_SCAN_LAST_DAYS = 3;
	private static final String TEMP_FILE_NAME_SUFFIX = ".tmp";

	private static final String SUBDIR_COMIC_STRIPS = "comic.strips";

	private static final String WEBSITE_TEMPLATE_ENCODING_NAME = "UTF-8";
	private static final String WEBSITE_TEMPLATE_PLACEHOLDER_TITLE = "<title>#TITLE#</title>";
	private static final String WEBSITE_TEMPLATE_PLACEHOLDER_STYLES = "<link rel=\"stylesheet\" type=\"text/css\" href=\"styles.css\" />";
	private static final String WEBSITE_TEMPLATE_PLACEHOLDER_TOC = "<p>#TABLE_OF_CONTENTS#</p>";
	private static final String WEBSITE_TEMPLATE_PLACEHOLDER_NAVIGATION = "<p>#NAVIGATION#</p>";
	private static final String WEBSITE_TEMPLATE_PLACEHOLDER_HEADING = "<p>#HEADING#</p>";
	private static final String WEBSITE_TEMPLATE_PLACEHOLDER_COMICS = "<p>#COMICS#</p>";

	private static final ThreadLocal<ByteBuffer> TL_BYTE_BUFFER = new ThreadLocal<ByteBuffer>() {
		@Override
		protected ByteBuffer initialValue() {
			return ByteBuffer.allocate( 64 * 1024 );
		}
	};

	private static final DateTimeFormatter FORMATTER_COMIC_STRIP_FILE_NAME = DateTimeFormatter.ofPattern( "'garfield.'uuuu-MM-dd-EEE'.gif'",
			LOCALE_DATE_TIME_FORMATTERS );
	private static final DateTimeFormatter FORMATTER_COMIC_STRIP_URL = DateTimeFormatter.ofPattern(
			"'https://d1ejxu6vysztl5.cloudfront.net/comics/garfield/'uuuu'/'uuuu-MM-dd'.gif'", LOCALE_DATE_TIME_FORMATTERS );
	//	private static final DateTimeFormatter FORMATTER_COMIC_STRIP_URL = DateTimeFormatter.ofPattern(
	//			"'http://images.ucomics.com/comics/ga/'uuuu'/ga'uuMMdd'.gif'", LOCALE_EN );
	private static final DateTimeFormatter FORMATTER_PAGE_MONTH_TITLE = DateTimeFormatter.ofPattern( "uuuu-MM MMMM", LOCALE_DATE_TIME_FORMATTERS );
	private static final DateTimeFormatter FORMATTER_PAGE_MONTH_TOC = DateTimeFormatter.ofPattern( "uuuu-MM MMM", LOCALE_DATE_TIME_FORMATTERS );
	private static final DateTimeFormatter FORMATTER_PAGE_MONTH_NAVIGATION = DateTimeFormatter.ofPattern( "uuuu-MM MMM", LOCALE_DATE_TIME_FORMATTERS );
	private static final DateTimeFormatter FORMATTER_PAGE_MONTH_FILE_NAME = DateTimeFormatter.ofPattern( "uuuu-MM'.xhtml'", LOCALE_DATE_TIME_FORMATTERS );
	private static final DateTimeFormatter FORMATTER_PAGE_YEAR_TOC = DateTimeFormatter.ofPattern( "uuuu", LOCALE_DATE_TIME_FORMATTERS );
	private static final DateTimeFormatter FORMATTER_PAGE_COMIC_STRIP_ALT_TEXT = DateTimeFormatter.ofPattern( "uuuu-MM-dd EEE", LOCALE_DATE_TIME_FORMATTERS );
	private static final DateTimeFormatter FORMATTER_PAGE_COMIC_STRIP_INDEX = DateTimeFormatter.ofPattern( "dd", LOCALE_DATE_TIME_FORMATTERS );

	private final Path targetDir;
	private final LocalDate greatestDateIncluding;

	private final Path subdirComics;

	public static void main( String[] args ) {

		if (args.length != 1) {
			System.out.println( APPLICATION_NAME + " " + VERSION );
			System.out.println( APPLICATION_DESCRIPTION );
			System.out.println();
			System.out.println( "Single argument required: Target directory" );
			System.exit( 1 );
		}
		System.out.println( "This is " + APPLICATION_NAME + " " + VERSION );
		new GetGarfield( Paths.get( args[0] ), getLatestComicStripDate() ).run();
	}

	public GetGarfield( Path targetDir, LocalDate greatestDateIncluding ) {
		super();

		this.targetDir = targetDir;
		this.greatestDateIncluding = greatestDateIncluding;
		this.subdirComics = targetDir.resolve( SUBDIR_COMIC_STRIPS );
	}

	@Override
	public void run() {

		System.out.println( "Target directory: \"" + targetDir + "\"" );
		System.out.println( "Getting comic strips up to including " + greatestDateIncluding );

		try {
			Files.createDirectories( subdirComics );
		} catch (IOException e) {
			throw new RuntimeException( "Failed to create target directory structure", e );
		}

		// Use a fixed thread pool that gets its jobs from a central queue
		// This causes the files to be downloaded in an approximately chronological order
		// With all the I/O going on queue contention is not an issue here
		ExecutorService executorService = Executors.newFixedThreadPool( CONCURRENT_DOWNLOADS );
		Runtime.getRuntime().addShutdownHook( createDownloadsShutdownHook( executorService ) );
		Iterator<LocalDate> daysToDownload = createStreamOfDays().filter( date -> !Files.exists( getComicFile( date ) ) ).iterator();
		// Start off by enqueueing as many download Runnables as there are concurrent download workers
		// plus a few more for good measure
		// After a Runnable has finished downloading its comic strip it will enqueue another Runnable
		// for the next comic strip that requires downloading
		// That way we avoid creating all Runnables up front, there are quite many of them for an
		// initial run
		for (int index = 0; index < (CONCURRENT_DOWNLOADS + 4); index++) {
			enqueueNextDownloadRunnable( executorService, daysToDownload );
		}
		System.out.println( "Started " + CONCURRENT_DOWNLOADS + " download worker thread(s)" );

		try { // Create some web pages while stuff is being downloaded
			createWebsite( "index.online", date -> FORMATTER_COMIC_STRIP_URL.format( date ) );
			createWebsite( "index", date -> "../" + SUBDIR_COMIC_STRIPS + "/" + FORMATTER_COMIC_STRIP_FILE_NAME.format( date ) );
		} catch (IOException e) {
			throw new RuntimeException( e );
		}

		// Wait until all downloads have completed
		try {
			while (!executorService.awaitTermination( Long.MAX_VALUE, TimeUnit.MILLISECONDS )) {
				// Nothing, just wait for termination again
			}
		} catch (@SuppressWarnings("unused") InterruptedException e) { // The downloads shutdown hook is handling this
			// System.out.println( "Downloading has been interrupted" );
			Thread.currentThread().interrupt();
		}
	}

	private static Thread createDownloadsShutdownHook( ExecutorService executorService ) {
		return new Thread( () -> {

			if (executorService.isTerminated()) {
				return; // Nothing to do
			}
			System.out.println( "Stopping downloads" );
			executorService.shutdownNow();
			System.out.println( "Waiting for running downloads to terminate" );
			try {
				while (!executorService.awaitTermination( Long.MAX_VALUE, TimeUnit.MILLISECONDS )) {
					// Nothing, just wait for termination again
				}
				System.out.println( "All running downloads have terminated" );
			} catch (@SuppressWarnings("unused") InterruptedException e) {
				Thread.currentThread().interrupt();
				System.out.println( "Some running downloads have been aborted, some temporary files may have been left behind" );
				// throw new RuntimeException( e );
			}
		}, "Downloads shutdown hook" );
	}

	public static LocalDate getLatestComicStripDate() {
		LocalDate latestUntestedDate = LocalDate.now( TIME_ZONE_COMIC_STRIP );
		LocalDate endDateExcluding = latestUntestedDate.plusDays( 1 );
		LocalDate result = null;

		// Including today, test the last n days to get a quick result if possible
		while ((ChronoUnit.DAYS.between( latestUntestedDate, endDateExcluding ) <= LATEST_COMIC_STRIP_SCAN_LAST_DAYS)
				&& (latestUntestedDate.compareTo( LATEST_COMIC_STRIP_DATE_KNOWN ) >= 0)) { // Only walk back to including latest-known-good
			if (isComicStripOnline( latestUntestedDate )) {
				// System.out.println( "Latest comic strip is from " + latestUntestedDate );
				return latestUntestedDate;
			}
			// Comic strip for this day is not online yet: Try one day earlier
			latestUntestedDate = latestUntestedDate.minusDays( 1 );
		}
		System.out.println( "No online comic strips found for the last " + LATEST_COMIC_STRIP_SCAN_LAST_DAYS + " days "
				+ "(" + latestUntestedDate.plusDays( 1 ) + ".." + endDateExcluding.minusDays( 1 ) + ")" );

		// Weird
		// Rule out e.g. connection issues or changed comic strip URLs before continuing with a broader search
		// Test a latest known good date
		if ((latestUntestedDate.plusDays( 1 ).equals( LATEST_COMIC_STRIP_DATE_KNOWN )) // I.e. latest-known-good was found to be offline in the preceding loop already
				|| !isComicStripOnline( LATEST_COMIC_STRIP_DATE_KNOWN )) {
			throw new RuntimeException( "Unable to determine most recent comic strip date: "
					+ "The known-to-exist comic strip for " + LATEST_COMIC_STRIP_DATE_KNOWN + " is unreachable, aborting" );
		}
		System.out.println( "Sanity check: Known-to-exist comic strip for " + LATEST_COMIC_STRIP_DATE_KNOWN + " is online" );

		// The comic strip appears to have been discontinued :(
		// Do a binary search in the range day-after-latest-known-good..latestUntestedDate, including
		if (ChronoUnit.DAYS.between( LATEST_COMIC_STRIP_DATE_KNOWN.plusDays( 1 ), latestUntestedDate.plusDays( 1 ) ) > 0) {
			System.out.println( "Binary searching for a newer latest comic strip in range "
					+ "" + LATEST_COMIC_STRIP_DATE_KNOWN.plusDays( 1 ) + ".." + latestUntestedDate + "" );
			result = getLatestComicStripDate( LATEST_COMIC_STRIP_DATE_KNOWN.plusDays( 1 ), latestUntestedDate.plusDays( 1 ) );
		}
		// The latest comic strip date is latest-known-good if the binary search did not yield anything newer
		result = (result == null) ? LATEST_COMIC_STRIP_DATE_KNOWN : result;
		// System.out.println( "Latest comic strip is from " + result );
		return result;
	}

	private static LocalDate getLatestComicStripDate( LocalDate startDateIncluding, LocalDate endDateExcluding ) {
		LocalDate startIncl = startDateIncluding;
		LocalDate endExcl = endDateExcluding;
		LocalDate result = null;
		LocalDate testing;

		while ((testing = getMiddle( startIncl, endExcl )) != null) {
			if (isComicStripOnline( testing )) {
				result = testing;
				startIncl = testing.plusDays( 1 );
			} else {
				endExcl = testing;
			}
		}
		return result;
	}

	private static LocalDate getMiddle( LocalDate startDateIncluding, LocalDate endDateExcluding ) {

		if (startDateIncluding.compareTo( endDateExcluding ) >= 0) {
			// Start date equal to or greater than end date
			return null;
		}
		return startDateIncluding.plusDays( ChronoUnit.DAYS.between( startDateIncluding, endDateExcluding ) / 2 );
	}

	private static boolean isComicStripOnline( LocalDate forDay ) {

		try (InputStream inputStream = new URL( FORMATTER_COMIC_STRIP_URL.format( forDay ) ).openStream()) {
			// System.out.println( forDay + " is online" );
			return true;
		} catch (@SuppressWarnings("unused") IOException e) {
			// System.out.println( forDay + " is offline" );
			return false;
		}
	}

	// TODO Use XML processing instead of line-based text search-and-replace
	private void createWebsite( String subdirName, Function<LocalDate, String> urlFactory ) throws IOException {

		Files.createDirectories( targetDir.resolve( subdirName ) );
		createStyles( subdirName );
		Iterator<LocalDate> months = createStreamOfMonths().iterator();
		while (months.hasNext()) {
			createMonthPage( subdirName, months.next(), urlFactory );
		}
		createIndexPage( subdirName );
		System.out.println( "Created web site: " + subdirName );
	}

	private void createStyles( String subdirName ) throws IOException {
		String stylesTemplateResourceName = "templates/styles.css";
		Path stylesFile = getStylesFile( subdirName );
		Path stylesFileTmp = targetDir.resolve( stylesFile.getFileName() + TEMP_FILE_NAME_SUFFIX );

		stylesFileTmp.toFile().deleteOnExit();
		try (ReadableByteChannel inputChannel = Channels.newChannel( GetGarfield.class.getResourceAsStream( stylesTemplateResourceName ) );
				SeekableByteChannel outputChannel = Files.newByteChannel( stylesFileTmp,
						StandardOpenOption.CREATE, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.WRITE )) {
			copy( inputChannel, outputChannel );
		}
		Files.move( stylesFileTmp, stylesFile, StandardCopyOption.REPLACE_EXISTING );
	}

	private Path getStylesFile( String subdirName ) {
		return targetDir.resolve( subdirName ).resolve( getStylesFileName() );
	}

	private String getStylesFileName() {
		return "styles.css";
	}

	private void createMonthPage( String subdirName, LocalDate month, Function<LocalDate, String> urlFactory ) throws IOException {
		LocalDate startOfMonth = month.withDayOfMonth( 1 );
		String monthTemplateResourceName = "templates/month.xhtml";
		Path monthFile = getMonthFile( subdirName, startOfMonth );
		Path monthFileTmp = targetDir.resolve( monthFile.getFileName() + TEMP_FILE_NAME_SUFFIX );
		boolean withAccessKeys = true;

		monthFileTmp.toFile().deleteOnExit();
		try (ReadableByteChannel inputChannel = Channels.newChannel( GetGarfield.class.getResourceAsStream( monthTemplateResourceName ) );
				BufferedReader reader = new BufferedReader( Channels.newReader( inputChannel, WEBSITE_TEMPLATE_ENCODING_NAME ) );
				SeekableByteChannel outputChannel = Files.newByteChannel( monthFileTmp,
						StandardOpenOption.CREATE, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.WRITE );
				BufferedWriter writer = new BufferedWriter( Channels.newWriter( outputChannel, WEBSITE_TEMPLATE_ENCODING_NAME ) )) {
			String line;

			while ((line = reader.readLine()) != null) {
				if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_TITLE )) {
					writeMonthTitle( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_TITLE ) ), startOfMonth );
				} else if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_STYLES )) {
					writeMonthStyles( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_STYLES ) ) );
				} else if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_NAVIGATION )) {
					writeMonthNavigation( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_NAVIGATION ) ),
							("../" + getIndexFileName( subdirName )), startOfMonth, withAccessKeys );
					withAccessKeys = false;
				} else if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_HEADING )) {
					writeMonthHeading( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_HEADING ) ), startOfMonth );
				} else if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_COMICS )) {
					writeMonthComicStrips( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_COMICS ) ), urlFactory, startOfMonth );
				} else {
					writer.write( line );
					writer.write( "\n" );
				}
			}
		}
		Files.move( monthFileTmp, monthFile, StandardCopyOption.REPLACE_EXISTING );
	}

	private void writeMonthTitle( BufferedWriter writer, String indentation, LocalDate forMonth ) throws IOException {
		writer.write( indentation + "<title>Garfield " + FORMATTER_PAGE_MONTH_TITLE.format( forMonth ) + "</title>" + "\n" );
	}

	private void writeMonthStyles( BufferedWriter writer, String indentation ) throws IOException {
		writer.write( indentation + "<link rel=\"stylesheet\" type=\"text/css\" href=\"" + getStylesFileName() + "\" />" + "\n" );
	}

	private void writeMonthNavigation( BufferedWriter writer, String indentation, String urlToIndex, LocalDate forMonth, boolean withAccessKeys )
			throws IOException {
		LocalDate startOfMonth = forMonth.withDayOfMonth( 1 );
		LocalDate previousMonth = startOfMonth.minusMonths( 1 );
		LocalDate nextMonth = startOfMonth.plusMonths( 1 );
		String textPreviousMonth = "←&nbsp;(a)&nbsp;" + FORMATTER_PAGE_MONTH_NAVIGATION.format( previousMonth );
		String textIndex = "Index&nbsp;(w)";
		String textNextMonth = FORMATTER_PAGE_MONTH_NAVIGATION.format( nextMonth ) + "&nbsp;(d)&nbsp;→";
		String indentation1 = indentation + "\t";

		writer.write( indentation + "<p>" + "\n" );
		writer.write( indentation1 );
		{
			if (previousMonth.compareTo( FIRST_COMIC_STRIP_DATE.withDayOfMonth( 1 ) ) >= 0) {
				writer.write( "<a href=\"" + getMonthFileName( previousMonth ) + "\"" );
				writer.write( withAccessKeys ? " accesskey=\"a\">" : ">" );
				writer.write( textPreviousMonth + "</a>" );
			} else {
				writer.write( textPreviousMonth );
			}
			writer.write( "&nbsp;" + "<a href=\"" + urlToIndex + "\"" );
			writer.write( withAccessKeys ? " accesskey=\"w\">" : ">" );
			writer.write( textIndex + "</a>" + "&nbsp;" );
			if (nextMonth.compareTo( greatestDateIncluding.withDayOfMonth( 1 ) ) <= 0) {
				writer.write( "<a href=\"" + getMonthFileName( nextMonth ) + "\"" );
				writer.write( withAccessKeys ? " accesskey=\"d\">" : ">" );
				writer.write( textNextMonth + "</a>" );
			} else {
				writer.write( textNextMonth );
			}
		}
		writer.write( "\n" );
		writer.write( indentation + "</p>" + "\n" );
	}

	private void writeMonthHeading( BufferedWriter writer, String indentation, LocalDate forMonth ) throws IOException {
		writer.write( indentation + "<h1>" + FORMATTER_PAGE_MONTH_TITLE.format( forMonth ) + "</h1>" + "\n" );
	}

	private void writeMonthComicStrips( BufferedWriter writer, String indentation, Function<LocalDate, String> urlFactory,
			LocalDate forMonth ) throws IOException {
		LocalDate startOfMonth = forMonth.withDayOfMonth( 1 );
		Iterator<LocalDate> days = createStreamOfDays( startOfMonth, startOfMonth.plusMonths( 1 ) ).iterator();

		while (days.hasNext()) {
			LocalDate day = days.next();
			if (day.compareTo( FIRST_COMIC_STRIP_DATE ) < 0) {
				continue;
			}
			if (day.compareTo( greatestDateIncluding ) > 0) {
				break;
			}
			writer.write( indentation + "<img" );
			writer.write( " src=\"" + urlFactory.apply( day ) + "\"" );
			writer.write( " id=\"" + FORMATTER_PAGE_COMIC_STRIP_INDEX.format( day ) + "\"" );
			writer.write( " alt=\"" + FORMATTER_PAGE_COMIC_STRIP_ALT_TEXT.format( day ) + "\"" );
			writer.write( " /><br />" + "\n" );
		}
	}

	private Path getMonthFile( String subdirName, LocalDate month ) {
		return targetDir.resolve( subdirName ).resolve( getMonthFileName( month ) );
	}

	private String getMonthFileName( LocalDate month ) {
		return FORMATTER_PAGE_MONTH_FILE_NAME.format( month );
	}

	private void createIndexPage( String subdirName ) throws IOException {
		String indexTemplateResourceName = "templates/index.xhtml";
		Path indexFile = getIndexFile( subdirName );
		Path indexFileTmp = targetDir.resolve( indexFile.getFileName() + TEMP_FILE_NAME_SUFFIX );

		indexFileTmp.toFile().deleteOnExit();
		try (ReadableByteChannel inputChannel = Channels.newChannel( GetGarfield.class.getResourceAsStream( indexTemplateResourceName ) );
				BufferedReader reader = new BufferedReader( Channels.newReader( inputChannel, WEBSITE_TEMPLATE_ENCODING_NAME ) );
				SeekableByteChannel outputChannel = Files.newByteChannel( indexFileTmp,
						StandardOpenOption.CREATE, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.WRITE );
				BufferedWriter writer = new BufferedWriter( Channels.newWriter( outputChannel, WEBSITE_TEMPLATE_ENCODING_NAME ) )) {
			String line;

			while ((line = reader.readLine()) != null) {
				if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_TITLE )) {
					writeIndexTitle( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_TITLE ) ) );
				} else if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_STYLES )) {
					writeIndexStyles( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_STYLES ) ), subdirName );
				} else if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_HEADING )) {
					writeIndexHeading( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_HEADING ) ) );
				} else if (line.endsWith( WEBSITE_TEMPLATE_PLACEHOLDER_TOC )) {
					writeIndexToc( writer, line.substring( 0, line.indexOf( WEBSITE_TEMPLATE_PLACEHOLDER_TOC ) ), subdirName );
				} else {
					writer.write( line );
					writer.write( "\n" );
				}
			}
		}
		Files.move( indexFileTmp, indexFile, StandardCopyOption.REPLACE_EXISTING );
	}

	private void writeIndexTitle( BufferedWriter writer, String indentation ) throws IOException {
		writer.write( indentation + "<title>Garfield</title>" + "\n" );
	}

	private void writeIndexStyles( BufferedWriter writer, String indentation, String subdirName ) throws IOException {
		writer.write( indentation + "<link rel=\"stylesheet\" type=\"text/css\" href=\"" + subdirName + "/" + getStylesFileName() + "\" />" + "\n" );
	}

	private void writeIndexHeading( BufferedWriter writer, String indentation ) throws IOException {
		writer.write( indentation + "<h1>Table of contents</h1>" + "\n" );
	}

	private void writeIndexToc( BufferedWriter writer, String indentation, String subdirName ) throws IOException {
		Iterator<LocalDate> years = createStreamOfYears().iterator();
		String indentation1 = indentation + "\t";

		while (years.hasNext()) {
			LocalDate year = years.next();
			writer.write( indentation + "<h2>" + FORMATTER_PAGE_YEAR_TOC.format( year ) + "</h2>" + "\n" );
			Iterator<LocalDate> months = createStreamOfMonths( year, year.plusYears( 1 ) ).iterator();
			writer.write( indentation + "<ul>" + "\n" );
			while (months.hasNext()) {
				LocalDate month = months.next();
				if (month.compareTo( FIRST_COMIC_STRIP_DATE.withDayOfMonth( 1 ) ) < 0) {
					continue;
				}
				if (month.compareTo( greatestDateIncluding.withDayOfMonth( 1 ) ) > 0) {
					break;
				}
				writer.write( indentation1 + "<li>" );
				writer.write( "<a href=\"" + subdirName + "/" + getMonthFileName( month ) + "\">" + FORMATTER_PAGE_MONTH_TOC.format( month ) + "</a>" );
				writer.write( "</li>" );
				writer.write( "\n" );
			}
			writer.write( indentation + "</ul>" + "\n" );
		}
	}

	private Path getIndexFile( String subdirName ) {
		return targetDir.resolve( getIndexFileName( subdirName ) );
	}

	private String getIndexFileName( String subdirName ) {
		return subdirName + ".xhtml";
	}

	private void enqueueNextDownloadRunnable( ExecutorService executorService, Iterator<LocalDate> daysToDownload ) {
		LocalDate dayToDownload = getNext( daysToDownload );

		if (dayToDownload == null) { // No more days to download the comic strip for
			executorService.shutdown();
			return;
		}
		if (executorService.isShutdown()) {
			return;
		}
		executorService.execute( createDownloadRunnable( executorService, daysToDownload, dayToDownload ) );
	}

	private Runnable createDownloadRunnable( ExecutorService executorService, Iterator<LocalDate> daysToDownload, LocalDate forDate ) {

		return () -> {
			downloadComicStrip( forDate );
			enqueueNextDownloadRunnable( executorService, daysToDownload );
		};
	}

	/**
	 * <p>
	 * Thread-safe method to get the next item from the given {@link Iterator}.
	 * </p>
	 *
	 * @param fromIterator
	 *            The {@link Iterator} from which to get the next item
	 * @return The {@link Iterator#next() next item} or <code>null</code> if the {@link Iterator} {@link Iterator#hasNext() does not have any more items}
	 */
	private static <T> T getNext( Iterator<T> fromIterator ) {

		synchronized (fromIterator) {
			return fromIterator.hasNext() ? fromIterator.next() : null;
		}
	}

	private Stream<LocalDate> createStreamOfYears() {
		return createStreamOfYears( FIRST_COMIC_STRIP_DATE, greatestDateIncluding.plusYears( 1 ) );
	}

	private Stream<LocalDate> createStreamOfMonths() {
		return createStreamOfMonths( FIRST_COMIC_STRIP_DATE, greatestDateIncluding.plusMonths( 1 ) );
	}

	private Stream<LocalDate> createStreamOfDays() {
		return createStreamOfDays( FIRST_COMIC_STRIP_DATE, greatestDateIncluding.plusDays( 1 ) );
	}

	private static Stream<LocalDate> createStreamOfYears( LocalDate fromIncluding, LocalDate toExcluding ) {
		LocalDate wholeYearFrom = fromIncluding.withDayOfYear( 1 );
		LocalDate wholeYearTo = toExcluding.withDayOfYear( 1 );
		long yearsCount = ChronoUnit.YEARS.between( wholeYearFrom, wholeYearTo );

		return Stream.iterate( wholeYearFrom, date -> date.plusYears( 1 ) ).limit( yearsCount );
	}

	private static Stream<LocalDate> createStreamOfMonths( LocalDate fromIncluding, LocalDate toExcluding ) {
		LocalDate wholeMonthFrom = fromIncluding.withDayOfMonth( 1 );
		LocalDate wholeMonthTo = toExcluding.withDayOfMonth( 1 );
		long monthsCount = ChronoUnit.MONTHS.between( wholeMonthFrom, wholeMonthTo );

		return Stream.iterate( wholeMonthFrom, date -> date.plusMonths( 1 ) ).limit( monthsCount );
	}

	private static Stream<LocalDate> createStreamOfDays( LocalDate fromIncluding, LocalDate toExcluding ) {
		long daysCount = ChronoUnit.DAYS.between( fromIncluding, toExcluding );

		return Stream.iterate( fromIncluding, date -> date.plusDays( 1 ) ).limit( daysCount );
	}

	private void downloadComicStrip( LocalDate forDate ) {
		Path targetFile = getComicFile( forDate );
		Path targetFileTmp = targetFile.getParent().resolve( targetFile.getFileName() + TEMP_FILE_NAME_SUFFIX );

		if (Files.exists( targetFile )) {
			return;
		}

		try {
			targetFileTmp.toFile().deleteOnExit();
			download( new URL( FORMATTER_COMIC_STRIP_URL.format( forDate ) ), targetFileTmp );
			Files.move( targetFileTmp, targetFile );
			System.out.println( "Downloaded " + FORMATTER_COMIC_STRIP_FILE_NAME.format( forDate ) );
		} catch (IOException e) {
			System.out.println( FORMATTER_COMIC_STRIP_FILE_NAME.format( forDate ) + ": Download failed (" + e.getClass().getSimpleName()
					+ ((e.getLocalizedMessage() != null) ? (": " + e.getLocalizedMessage()) : "") + ")" );
		}
	}

	private void download( URL url, Path toFile ) throws IOException {

		try (ReadableByteChannel inputChannel = Channels.newChannel( url.openStream() );
				SeekableByteChannel outputChannel = Files.newByteChannel( toFile,
						StandardOpenOption.CREATE, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.WRITE )) {
			copy( inputChannel, outputChannel );
		}
	}

	private static void copy( ReadableByteChannel inputChannel, SeekableByteChannel outputChannel ) throws IOException {
		ByteBuffer buffer = TL_BYTE_BUFFER.get();

		buffer.clear();
		while (inputChannel.read( buffer ) >= 0) {
			buffer.flip();
			while (buffer.position() < buffer.limit()) {
				outputChannel.write( buffer );
			}
			buffer.clear();
		}
	}

	private Path getComicFile( LocalDate forDate ) {
		return subdirComics.resolve( FORMATTER_COMIC_STRIP_FILE_NAME.format( forDate ) );
	}
}
